const NO_OP = '$NO_OP';
const ERROR_MSG = 'a runtime error occured! Use Inferno in development environment to find the error.';
const isArray = Array.isArray;
function isStringOrNumber(o) {
  const type = typeof o;
  return type === 'string' || type === 'number';
}
function isNullOrUndef(o) {
  return isUndefined(o) || isNull(o);
}
function isInvalid(o) {
  return isNull(o) || o === false || isTrue(o) || isUndefined(o);
}
function isFunction(o) {
  return typeof o === 'function';
}
function isString(o) {
  return typeof o === 'string';
}
function isNumber(o) {
  return typeof o === 'number';
}
function isNull(o) {
  return o === null;
}
function isTrue(o) {
  return o === true;
}
function isUndefined(o) {
  return o === void 0;
}
function throwError(message) {
  if (!message) {
    message = ERROR_MSG;
  }
  throw new Error(`Inferno Error: ${message}`);
}
function combineFrom(first, second) {
  const out = {};
  if (first) {
    for (const key in first) {
      out[key] = first[key];
    }
  }
  if (second) {
    for (const key in second) {
      out[key] = second[key];
    }
  }
  return out;
}

const keyPrefix = '$';
function getVNode(childFlags, children, className, flags, key, props, ref, type) {
  return {
    childFlags,
    children,
    className,
    dom: null,
    flags,
    key: key === void 0 ? null : key,
    parentVNode: null,
    props: props === void 0 ? null : props,
    ref: ref === void 0 ? null : ref,
    type
  };
}
function createVNode(flags, type, className, children, childFlags, props, key, ref) {
  const childFlag = childFlags === void 0 ? 1 /* ChildFlags.HasInvalidChildren */ : childFlags;
  const vNode = getVNode(childFlag, children, className, flags, key, props, ref, type);
  const optsVNode = options.createVNode;
  if (typeof optsVNode === 'function') {
    optsVNode(vNode);
  }
  if (childFlag === 0 /* ChildFlags.UnknownChildren */) {
    normalizeChildren(vNode, vNode.children);
  }
  return vNode;
}
function createComponentVNode(flags, type, props, key, ref) {
  if ((flags & 2 /* VNodeFlags.ComponentUnknown */) > 0) {
    flags = type.prototype && isFunction(type.prototype.render) ? 4 /* VNodeFlags.ComponentClass */ : 8 /* VNodeFlags.ComponentFunction */;
  }
  // set default props
  const defaultProps = type.defaultProps;
  if (!isNullOrUndef(defaultProps)) {
    if (!props) {
      props = {}; // Props can be referenced and modified at application level so always create new object
    }
    for (const prop in defaultProps) {
      if (isUndefined(props[prop])) {
        props[prop] = defaultProps[prop];
      }
    }
  }
  if ((flags & 8 /* VNodeFlags.ComponentFunction */) > 0) {
    const defaultHooks = type.defaultHooks;
    if (!isNullOrUndef(defaultHooks)) {
      if (!ref) {
        // As ref cannot be referenced from application level, we can use the same refs object
        ref = defaultHooks;
      } else {
        for (const prop in defaultHooks) {
          if (isUndefined(ref[prop])) {
            ref[prop] = defaultHooks[prop];
          }
        }
      }
    }
  }
  const vNode = getVNode(1 /* ChildFlags.HasInvalidChildren */, null, null, flags, key, props, ref, type);
  const optsVNode = options.createVNode;
  if (isFunction(optsVNode)) {
    optsVNode(vNode);
  }
  return vNode;
}
function createTextVNode(text, key) {
  return getVNode(1 /* ChildFlags.HasInvalidChildren */, isNullOrUndef(text) ? '' : text, null, 16 /* VNodeFlags.Text */, key, null, null, null);
}
function normalizeProps(vNode) {
  const props = vNode.props;
  if (props) {
    const flags = vNode.flags;
    if (flags & 481 /* VNodeFlags.Element */) {
      if (props.children !== void 0 && isNullOrUndef(vNode.children)) {
        normalizeChildren(vNode, props.children);
      }
      if (props.className !== void 0) {
        vNode.className = props.className || null;
        props.className = undefined;
      }
    }
    if (props.key !== void 0) {
      vNode.key = props.key;
      props.key = undefined;
    }
    if (props.ref !== void 0) {
      if (flags & 8 /* VNodeFlags.ComponentFunction */) {
        vNode.ref = combineFrom(vNode.ref, props.ref);
      } else {
        vNode.ref = props.ref;
      }
      props.ref = undefined;
    }
  }
  return vNode;
}
function directClone(vNodeToClone) {
  let newVNode;
  const flags = vNodeToClone.flags;
  if (flags & 14 /* VNodeFlags.Component */) {
    let props;
    const propsToClone = vNodeToClone.props;
    if (!isNull(propsToClone)) {
      props = {};
      for (const key in propsToClone) {
        props[key] = propsToClone[key];
      }
    }
    newVNode = createComponentVNode(flags, vNodeToClone.type, props, vNodeToClone.key, vNodeToClone.ref);
  } else if (flags & 481 /* VNodeFlags.Element */) {
    newVNode = createVNode(flags, vNodeToClone.type, vNodeToClone.className, vNodeToClone.children, vNodeToClone.childFlags, vNodeToClone.props, vNodeToClone.key, vNodeToClone.ref);
  } else if (flags & 16 /* VNodeFlags.Text */) {
    newVNode = createTextVNode(vNodeToClone.children, vNodeToClone.key);
  } else if (flags & 1024 /* VNodeFlags.Portal */) {
    newVNode = vNodeToClone;
  }
  return newVNode;
}
function createVoidVNode() {
  return createTextVNode('', null);
}
function _normalizeVNodes(nodes, result, index, currentKey) {
  for (const len = nodes.length; index < len; index++) {
    let n = nodes[index];
    if (!isInvalid(n)) {
      const newKey = currentKey + keyPrefix + index;
      if (isArray(n)) {
        _normalizeVNodes(n, result, 0, newKey);
      } else {
        if (isStringOrNumber(n)) {
          n = createTextVNode(n, newKey);
        } else {
          const oldKey = n.key;
          const isPrefixedKey = isString(oldKey) && oldKey[0] === keyPrefix;
          if (!isNull(n.dom) || isPrefixedKey) {
            n = directClone(n);
          }
          if (isNull(oldKey) || isPrefixedKey) {
            n.key = newKey;
          } else {
            n.key = currentKey + oldKey;
          }
        }
        result.push(n);
      }
    }
  }
}
function getFlagsForElementVnode(type) {
  if (type === 'svg') {
    return 32 /* VNodeFlags.SvgElement */;
  }
  if (type === 'input') {
    return 64 /* VNodeFlags.InputElement */;
  }
  if (type === 'select') {
    return 256 /* VNodeFlags.SelectElement */;
  }
  if (type === 'textarea') {
    return 128 /* VNodeFlags.TextareaElement */;
  }
  return 1 /* VNodeFlags.HtmlElement */;
}
function normalizeChildren(vNode, children) {
  let newChildren;
  let newChildFlags = 1 /* ChildFlags.HasInvalidChildren */;
  // Don't change children to match strict equal (===) true in patching
  if (isInvalid(children)) {
    newChildren = children;
  } else if (isString(children)) {
    newChildFlags = 2 /* ChildFlags.HasVNodeChildren */;
    newChildren = createTextVNode(children);
  } else if (isNumber(children)) {
    newChildFlags = 2 /* ChildFlags.HasVNodeChildren */;
    newChildren = createTextVNode(children + '');
  } else if (isArray(children)) {
    const len = children.length;
    if (len === 0) {
      newChildren = null;
      newChildFlags = 1 /* ChildFlags.HasInvalidChildren */;
    } else {
      // we assign $ which basically means we've flagged this array for future note
      // if it comes back again, we need to clone it, as people are using it
      // in an immutable way
      // tslint:disable-next-line
      if (Object.isFrozen(children) || children['$'] === true) {
        children = children.slice();
      }
      newChildFlags = 8 /* ChildFlags.HasKeyedChildren */;
      for (let i = 0; i < len; i++) {
        let n = children[i];
        if (isInvalid(n) || isArray(n)) {
          newChildren = newChildren || children.slice(0, i);
          _normalizeVNodes(children, newChildren, i, '');
          break;
        } else if (isStringOrNumber(n)) {
          newChildren = newChildren || children.slice(0, i);
          newChildren.push(createTextVNode(n, keyPrefix + i));
        } else {
          const key = n.key;
          const isNullDom = isNull(n.dom);
          const isNullKey = isNull(key);
          const isPrefixed = !isNullKey && isString(key) && key[0] === keyPrefix;
          if (!isNullDom || isNullKey || isPrefixed) {
            newChildren = newChildren || children.slice(0, i);
            if (!isNullDom || isPrefixed) {
              n = directClone(n);
            }
            if (isNullKey || isPrefixed) {
              n.key = keyPrefix + i;
            }
            newChildren.push(n);
          } else if (newChildren) {
            newChildren.push(n);
          }
        }
      }
      newChildren = newChildren || children;
      newChildren.$ = true;
    }
  } else {
    newChildren = children;
    if (!isNull(children.dom)) {
      newChildren = directClone(children);
    }
    newChildFlags = 2 /* ChildFlags.HasVNodeChildren */;
  }
  vNode.children = newChildren;
  vNode.childFlags = newChildFlags;
  return vNode;
}
const options = {
  afterRender: null,
  beforeRender: null,
  createVNode: null,
  renderComplete: null
};

/**
 * Links given data to event as first parameter
 * @param {*} data data to be linked, it will be available in function as first parameter
 * @param {Function} event Function to be called when event occurs
 * @returns {{data: *, event: Function}}
 */
function linkEvent(data, event) {
  if (isFunction(event)) {
    return {
      data,
      event
    };
  }
  return null; // Return null when event is invalid, to avoid creating unnecessary event handlers
}

const xlinkNS = 'http://www.w3.org/1999/xlink';
const xmlNS = 'http://www.w3.org/XML/1998/namespace';
const svgNS = 'http://www.w3.org/2000/svg';
const namespaces = {
  'xlink:actuate': xlinkNS,
  'xlink:arcrole': xlinkNS,
  'xlink:href': xlinkNS,
  'xlink:role': xlinkNS,
  'xlink:show': xlinkNS,
  'xlink:title': xlinkNS,
  'xlink:type': xlinkNS,
  'xml:base': xmlNS,
  'xml:lang': xmlNS,
  'xml:space': xmlNS
};

// We need EMPTY_OBJ defined in one place.
// Its used for comparison so we cant inline it into shared
const EMPTY_OBJ = {};
const LIFECYCLE = [];
function appendChild(parentDom, dom) {
  parentDom.appendChild(dom);
}
function insertOrAppend(parentDom, newNode, nextNode) {
  if (isNullOrUndef(nextNode)) {
    appendChild(parentDom, newNode);
  } else {
    parentDom.insertBefore(newNode, nextNode);
  }
}
function documentCreateElement(tag, isSVG) {
  if (isSVG) {
    return document.createElementNS(svgNS, tag);
  }
  return document.createElement(tag);
}
function replaceChild(parentDom, newDom, lastDom) {
  parentDom.replaceChild(newDom, lastDom);
}
function removeChild(parentDom, dom) {
  parentDom.removeChild(dom);
}
function callAll(arrayFn) {
  let listener;
  while ((listener = arrayFn.shift()) !== undefined) {
    listener();
  }
}

const attachedEventCounts = {};
const attachedEvents = {};
function handleEvent(name, nextEvent, dom) {
  const eventsLeft = attachedEventCounts[name];
  let eventsObject = dom.$EV;
  if (nextEvent) {
    if (!eventsLeft) {
      attachedEvents[name] = attachEventToDocument(name);
      attachedEventCounts[name] = 0;
    }
    if (!eventsObject) {
      eventsObject = dom.$EV = {};
    }
    if (!eventsObject[name]) {
      attachedEventCounts[name]++;
    }
    eventsObject[name] = nextEvent;
  } else if (eventsObject && eventsObject[name]) {
    attachedEventCounts[name]--;
    if (eventsLeft === 1) {
      document.removeEventListener(normalizeEventName(name), attachedEvents[name]);
      attachedEvents[name] = null;
    }
    eventsObject[name] = nextEvent;
  }
}
// When browsers fully support event.composedPath we could loop it through instead of using parentNode property
function getTargetNode(event) {
  return isFunction(event.composedPath) ? event.composedPath()[0] : event.target;
}
function dispatchEvents(event, isClick, name, eventData) {
  let dom = getTargetNode(event);
  while (!isNull(dom)) {
    // Html Nodes can be nested fe: span inside button in that scenario browser does not handle disabled attribute on parent,
    // because the event listener is on document.body
    // Don't process clicks on disabled elements
    if (isClick && dom.disabled) {
      return;
    }
    const eventsObject = dom.$EV;
    if (eventsObject) {
      const currentEvent = eventsObject[name];
      if (currentEvent) {
        // linkEvent object
        eventData.dom = dom;
        if (currentEvent.event) {
          currentEvent.event(currentEvent.data, event);
        } else {
          currentEvent(event);
        }
        if (event.cancelBubble) {
          return;
        }
      }
    }
    dom = dom.parentNode;
  }
}
function normalizeEventName(name) {
  return name.substr(2).toLowerCase();
}
function stopPropagation() {
  this.cancelBubble = true;
  if (!this.immediatePropagationStopped) {
    this.stopImmediatePropagation();
  }
}
function attachEventToDocument(name) {
  const docEvent = event => {
    const type = event.type;
    const isClick = type === 'click' || type === 'dblclick';
    if (isClick && event.button !== 0) {
      // Firefox incorrectly triggers click event for mid/right mouse buttons.
      // This bug has been active for 12 years.
      // https://bugzilla.mozilla.org/show_bug.cgi?id=184051
      event.stopPropagation();
      return false;
    }
    event.stopPropagation = stopPropagation;
    // Event data needs to be object to save reference to currentTarget getter
    const eventData = {
      dom: document
    };
    Object.defineProperty(event, 'currentTarget', {
      configurable: true,
      get: function get() {
        return eventData.dom;
      }
    });
    dispatchEvents(event, isClick, name, eventData);
    return;
  };
  document.addEventListener(normalizeEventName(name), docEvent);
  return docEvent;
}

function isSameInnerHTML(dom, innerHTML) {
  const tempdom = document.createElement('i');
  tempdom.innerHTML = innerHTML;
  return tempdom.innerHTML === dom.innerHTML;
}
function isSamePropsInnerHTML(dom, props) {
  return Boolean(props && props.dangerouslySetInnerHTML && props.dangerouslySetInnerHTML.__html && isSameInnerHTML(dom, props.dangerouslySetInnerHTML.__html));
}

function triggerEventListener(props, methodName, e) {
  if (props[methodName]) {
    const listener = props[methodName];
    if (listener.event) {
      listener.event(listener.data, e);
    } else {
      listener(e);
    }
  } else {
    const nativeListenerName = methodName.toLowerCase();
    if (props[nativeListenerName]) {
      props[nativeListenerName](e);
    }
  }
}
function createWrappedFunction(methodName, applyValue) {
  const fnMethod = function (e) {
    e.stopPropagation();
    const vNode = this.$V;
    // If vNode is gone by the time event fires, no-op
    if (!vNode) {
      return;
    }
    const props = vNode.props || EMPTY_OBJ;
    const dom = vNode.dom;
    if (isString(methodName)) {
      triggerEventListener(props, methodName, e);
    } else {
      for (let i = 0; i < methodName.length; i++) {
        triggerEventListener(props, methodName[i], e);
      }
    }
    if (isFunction(applyValue)) {
      const newVNode = this.$V;
      const newProps = newVNode.props || EMPTY_OBJ;
      applyValue(newProps, dom, false, newVNode);
    }
  };
  Object.defineProperty(fnMethod, 'wrapped', {
    configurable: false,
    enumerable: false,
    value: true,
    writable: false
  });
  return fnMethod;
}

function isCheckedType(type) {
  return type === 'checkbox' || type === 'radio';
}
const onTextInputChange = createWrappedFunction('onInput', applyValueInput);
const wrappedOnChange$1 = createWrappedFunction(['onClick', 'onChange'], applyValueInput);
/* tslint:disable-next-line:no-empty */
function emptywrapper(event) {
  event.stopPropagation();
}
emptywrapper.wrapped = true;
function inputEvents(dom, nextPropsOrEmpty) {
  if (isCheckedType(nextPropsOrEmpty.type)) {
    dom.onchange = wrappedOnChange$1;
    dom.onclick = emptywrapper;
  } else {
    dom.oninput = onTextInputChange;
  }
}
function applyValueInput(nextPropsOrEmpty, dom) {
  const type = nextPropsOrEmpty.type;
  const value = nextPropsOrEmpty.value;
  const checked = nextPropsOrEmpty.checked;
  const multiple = nextPropsOrEmpty.multiple;
  const defaultValue = nextPropsOrEmpty.defaultValue;
  const hasValue = !isNullOrUndef(value);
  if (type && type !== dom.type) {
    dom.setAttribute('type', type);
  }
  if (!isNullOrUndef(multiple) && multiple !== dom.multiple) {
    dom.multiple = multiple;
  }
  if (!isNullOrUndef(defaultValue) && !hasValue) {
    dom.defaultValue = defaultValue + '';
  }
  if (isCheckedType(type)) {
    if (hasValue) {
      dom.value = value;
    }
    if (!isNullOrUndef(checked)) {
      dom.checked = checked;
    }
  } else {
    if (hasValue && dom.value !== value) {
      dom.defaultValue = value;
      dom.value = value;
    } else if (!isNullOrUndef(checked)) {
      dom.checked = checked;
    }
  }
}

function updateChildOptionGroup(vNode, value) {
  const type = vNode.type;
  if (type === 'optgroup') {
    const children = vNode.children;
    const childFlags = vNode.childFlags;
    if (childFlags & 12 /* ChildFlags.MultipleChildren */) {
      for (let i = 0, len = children.length; i < len; i++) {
        updateChildOption(children[i], value);
      }
    } else if (childFlags === 2 /* ChildFlags.HasVNodeChildren */) {
      updateChildOption(children, value);
    }
  } else {
    updateChildOption(vNode, value);
  }
}
function updateChildOption(vNode, value) {
  const props = vNode.props || EMPTY_OBJ;
  const dom = vNode.dom;
  // we do this as multiple may have changed
  dom.value = props.value;
  if (isArray(value) && value.indexOf(props.value) !== -1 || props.value === value) {
    dom.selected = true;
  } else if (!isNullOrUndef(value) || !isNullOrUndef(props.selected)) {
    dom.selected = props.selected || false;
  }
}
const onSelectChange = createWrappedFunction('onChange', applyValueSelect);
function selectEvents(dom) {
  dom.onchange = onSelectChange;
}
function applyValueSelect(nextPropsOrEmpty, dom, mounting, vNode) {
  const multiplePropInBoolean = Boolean(nextPropsOrEmpty.multiple);
  if (!isNullOrUndef(nextPropsOrEmpty.multiple) && multiplePropInBoolean !== dom.multiple) {
    dom.multiple = multiplePropInBoolean;
  }
  const childFlags = vNode.childFlags;
  if ((childFlags & 1 /* ChildFlags.HasInvalidChildren */) === 0) {
    const children = vNode.children;
    let value = nextPropsOrEmpty.value;
    if (mounting && isNullOrUndef(value)) {
      value = nextPropsOrEmpty.defaultValue;
    }
    if (childFlags & 12 /* ChildFlags.MultipleChildren */) {
      for (let i = 0, len = children.length; i < len; i++) {
        updateChildOptionGroup(children[i], value);
      }
    } else if (childFlags === 2 /* ChildFlags.HasVNodeChildren */) {
      updateChildOptionGroup(children, value);
    }
  }
}

const onTextareaInputChange = createWrappedFunction('onInput', applyValueTextArea);
const wrappedOnChange = createWrappedFunction('onChange');
function textAreaEvents(dom, nextPropsOrEmpty) {
  dom.oninput = onTextareaInputChange;
  if (nextPropsOrEmpty.onChange) {
    dom.onchange = wrappedOnChange;
  }
}
function applyValueTextArea(nextPropsOrEmpty, dom, mounting) {
  const value = nextPropsOrEmpty.value;
  const domValue = dom.value;
  if (isNullOrUndef(value)) {
    if (mounting) {
      const defaultValue = nextPropsOrEmpty.defaultValue;
      if (!isNullOrUndef(defaultValue) && defaultValue !== domValue) {
        dom.defaultValue = defaultValue;
        dom.value = defaultValue;
      }
    }
  } else if (domValue !== value) {
    /* There is value so keep it controlled */
    dom.defaultValue = value;
    dom.value = value;
  }
}

/**
 * There is currently no support for switching same input between controlled and nonControlled
 * If that ever becomes a real issue, then re design controlled elements
 * Currently user must choose either controlled or non-controlled and stick with that
 */
function processElement(flags, vNode, dom, nextPropsOrEmpty, mounting, isControlled) {
  if (flags & 64 /* VNodeFlags.InputElement */) {
    applyValueInput(nextPropsOrEmpty, dom);
  } else if (flags & 256 /* VNodeFlags.SelectElement */) {
    applyValueSelect(nextPropsOrEmpty, dom, mounting, vNode);
  } else if (flags & 128 /* VNodeFlags.TextareaElement */) {
    applyValueTextArea(nextPropsOrEmpty, dom, mounting);
  }
  if (isControlled) {
    dom.$V = vNode;
  }
}
function addFormElementEventHandlers(flags, dom, nextPropsOrEmpty) {
  if (flags & 64 /* VNodeFlags.InputElement */) {
    inputEvents(dom, nextPropsOrEmpty);
  } else if (flags & 256 /* VNodeFlags.SelectElement */) {
    selectEvents(dom);
  } else if (flags & 128 /* VNodeFlags.TextareaElement */) {
    textAreaEvents(dom, nextPropsOrEmpty);
  }
}
function isControlledFormElement(nextPropsOrEmpty) {
  return nextPropsOrEmpty.type && isCheckedType(nextPropsOrEmpty.type) ? !isNullOrUndef(nextPropsOrEmpty.checked) : !isNullOrUndef(nextPropsOrEmpty.value);
}

function remove(vNode, parentDom) {
  unmount(vNode);
  if (parentDom && vNode.dom) {
    removeChild(parentDom, vNode.dom);
    // Let carbage collector free memory
    vNode.dom = null;
  }
}
function unmount(vNode) {
  const flags = vNode.flags;
  if (flags & 481 /* VNodeFlags.Element */) {
    const ref = vNode.ref;
    const props = vNode.props;
    if (isFunction(ref)) {
      ref(null);
    }
    const children = vNode.children;
    const childFlags = vNode.childFlags;
    if (childFlags & 12 /* ChildFlags.MultipleChildren */) {
      unmountAllChildren(children);
    } else if (childFlags === 2 /* ChildFlags.HasVNodeChildren */) {
      unmount(children);
    }
    if (!isNull(props)) {
      for (const name in props) {
        switch (name) {
          case 'onClick':
          case 'onDblClick':
          case 'onFocusIn':
          case 'onFocusOut':
          case 'onKeyDown':
          case 'onKeyPress':
          case 'onKeyUp':
          case 'onMouseDown':
          case 'onMouseMove':
          case 'onMouseUp':
          case 'onSubmit':
          case 'onTouchEnd':
          case 'onTouchMove':
          case 'onTouchStart':
            handleEvent(name, null, vNode.dom);
            break;
        }
      }
    }
  } else {
    const children = vNode.children;
    // Safe guard for crashed VNode
    if (children) {
      if (flags & 14 /* VNodeFlags.Component */) {
        const ref = vNode.ref;
        if (flags & 4 /* VNodeFlags.ComponentClass */) {
          if (isFunction(children.componentWillUnmount)) {
            children.componentWillUnmount();
          }
          if (isFunction(ref)) {
            ref(null);
          }
          children.$UN = true;
          if (children.$LI) {
            unmount(children.$LI);
          }
        } else {
          if (!isNullOrUndef(ref) && isFunction(ref.onComponentWillUnmount)) {
            ref.onComponentWillUnmount(vNode.dom, vNode.props || EMPTY_OBJ);
          }
          unmount(children);
        }
      } else if (flags & 1024 /* VNodeFlags.Portal */) {
        remove(children, vNode.type);
      }
    }
  }
}
function unmountAllChildren(children) {
  for (let i = 0, len = children.length; i < len; i++) {
    unmount(children[i]);
  }
}
function removeAllChildren(dom, children) {
  unmountAllChildren(children);
  dom.textContent = '';
}

function createLinkEvent(linkEvent, nextValue) {
  return function (e) {
    linkEvent(nextValue.data, e);
  };
}
function patchEvent(name, nextValue, dom) {
  const nameLowerCase = name.toLowerCase();
  if (!isFunction(nextValue) && !isNullOrUndef(nextValue)) {
    const linkEvent = nextValue.event;
    if (linkEvent && isFunction(linkEvent)) {
      dom[nameLowerCase] = createLinkEvent(linkEvent, nextValue);
    }
  } else {
    const domEvent = dom[nameLowerCase];
    // if the function is wrapped, that means it's been controlled by a wrapper
    if (!domEvent || !domEvent.wrapped) {
      dom[nameLowerCase] = nextValue;
    }
  }
}
function getNumberStyleValue(style, value) {
  switch (style) {
    case 'animationIterationCount':
    case 'borderImageOutset':
    case 'borderImageSlice':
    case 'borderImageWidth':
    case 'boxFlex':
    case 'boxFlexGroup':
    case 'boxOrdinalGroup':
    case 'columnCount':
    case 'fillOpacity':
    case 'flex':
    case 'flexGrow':
    case 'flexNegative':
    case 'flexOrder':
    case 'flexPositive':
    case 'flexShrink':
    case 'floodOpacity':
    case 'fontWeight':
    case 'gridColumn':
    case 'gridRow':
    case 'lineClamp':
    case 'lineHeight':
    case 'opacity':
    case 'order':
    case 'orphans':
    case 'stopOpacity':
    case 'strokeDasharray':
    case 'strokeDashoffset':
    case 'strokeMiterlimit':
    case 'strokeOpacity':
    case 'strokeWidth':
    case 'tabSize':
    case 'widows':
    case 'zIndex':
    case 'zoom':
      return value;
    default:
      return value + 'px';
  }
}
// We are assuming here that we come from patchProp routine
// -nextAttrValue cannot be null or undefined
function patchStyle(lastAttrValue, nextAttrValue, dom) {
  const domStyle = dom.style;
  let style;
  let value;
  if (isString(nextAttrValue)) {
    domStyle.cssText = nextAttrValue;
    return;
  }
  if (!isNullOrUndef(lastAttrValue) && !isString(lastAttrValue)) {
    for (style in nextAttrValue) {
      // do not add a hasOwnProperty check here, it affects performance
      value = nextAttrValue[style];
      if (value !== lastAttrValue[style]) {
        domStyle[style] = isNumber(value) ? getNumberStyleValue(style, value) : value;
      }
    }
    for (style in lastAttrValue) {
      if (isNullOrUndef(nextAttrValue[style])) {
        domStyle[style] = '';
      }
    }
  } else {
    for (style in nextAttrValue) {
      value = nextAttrValue[style];
      domStyle[style] = isNumber(value) ? getNumberStyleValue(style, value) : value;
    }
  }
}
function patchProp(prop, lastValue, nextValue, dom, isSVG, hasControlledValue, lastVNode) {
  switch (prop) {
    case 'onClick':
    case 'onDblClick':
    case 'onFocusIn':
    case 'onFocusOut':
    case 'onKeyDown':
    case 'onKeyPress':
    case 'onKeyUp':
    case 'onMouseDown':
    case 'onMouseMove':
    case 'onMouseUp':
    case 'onSubmit':
    case 'onTouchEnd':
    case 'onTouchMove':
    case 'onTouchStart':
      handleEvent(prop, nextValue, dom);
      break;
    case 'children':
    case 'childrenType':
    case 'className':
    case 'defaultValue':
    case 'key':
    case 'multiple':
    case 'ref':
      break;
    case 'autoFocus':
      dom.autofocus = !!nextValue;
      break;
    case 'allowfullscreen':
    case 'autoplay':
    case 'capture':
    case 'checked':
    case 'controls':
    case 'default':
    case 'disabled':
    case 'hidden':
    case 'indeterminate':
    case 'loop':
    case 'muted':
    case 'novalidate':
    case 'open':
    case 'readOnly':
    case 'required':
    case 'reversed':
    case 'scoped':
    case 'seamless':
    case 'selected':
      dom[prop] = !!nextValue;
      break;
    case 'defaultChecked':
    case 'value':
    case 'volume':
      if (hasControlledValue && prop === 'value') {
        return;
      }
      const value = isNullOrUndef(nextValue) ? '' : nextValue;
      if (dom[prop] !== value) {
        dom[prop] = value;
      }
      break;
    case 'dangerouslySetInnerHTML':
      const lastHtml = lastValue && lastValue.__html || '';
      const nextHtml = nextValue && nextValue.__html || '';
      if (lastHtml !== nextHtml) {
        if (!isNullOrUndef(nextHtml) && !isSameInnerHTML(dom, nextHtml)) {
          if (!isNull(lastVNode)) {
            if (lastVNode.childFlags & 12 /* ChildFlags.MultipleChildren */) {
              unmountAllChildren(lastVNode.children);
            } else if (lastVNode.childFlags === 2 /* ChildFlags.HasVNodeChildren */) {
              unmount(lastVNode.children);
            }
            lastVNode.children = null;
            lastVNode.childFlags = 1 /* ChildFlags.HasInvalidChildren */;
          }
          dom.innerHTML = nextHtml;
        }
      }
      break;
    default:
      if (prop[0] === 'o' && prop[1] === 'n') {
        patchEvent(prop, nextValue, dom);
      } else if (isNullOrUndef(nextValue)) {
        dom.removeAttribute(prop);
      } else if (prop === 'style') {
        patchStyle(lastValue, nextValue, dom);
      } else if (isSVG && namespaces[prop]) {
        // We optimize for isSVG being false
        // If we end up in this path we can read property again
        dom.setAttributeNS(namespaces[prop], prop, nextValue);
      } else {
        dom.setAttribute(prop, nextValue);
      }
      break;
  }
}
function mountProps(vNode, flags, props, dom, isSVG) {
  let hasControlledValue = false;
  const isFormElement = (flags & 448 /* VNodeFlags.FormElement */) > 0;
  if (isFormElement) {
    hasControlledValue = isControlledFormElement(props);
    if (hasControlledValue) {
      addFormElementEventHandlers(flags, dom, props);
    }
  }
  for (const prop in props) {
    // do not add a hasOwnProperty check here, it affects performance
    patchProp(prop, null, props[prop], dom, isSVG, hasControlledValue, null);
  }
  if (isFormElement) {
    processElement(flags, vNode, dom, props, true, hasControlledValue);
  }
}

function createClassComponentInstance(vNode, Component, props, context) {
  const instance = new Component(props, context);
  vNode.children = instance;
  instance.$V = vNode;
  instance.$BS = false;
  instance.context = context;
  if (instance.props === EMPTY_OBJ) {
    instance.props = props;
  }
  instance.$UN = false;
  if (isFunction(instance.componentWillMount)) {
    instance.$BR = true;
    instance.componentWillMount();
    if (instance.$PSS) {
      const state = instance.state;
      const pending = instance.$PS;
      if (isNull(state)) {
        instance.state = pending;
      } else {
        for (const key in pending) {
          state[key] = pending[key];
        }
      }
      instance.$PSS = false;
      instance.$PS = null;
    }
    instance.$BR = false;
  }
  if (isFunction(options.beforeRender)) {
    options.beforeRender(instance);
  }
  const input = handleComponentInput(instance.render(props, instance.state, context), vNode);
  let childContext;
  if (isFunction(instance.getChildContext)) {
    childContext = instance.getChildContext();
  }
  if (isNullOrUndef(childContext)) {
    instance.$CX = context;
  } else {
    instance.$CX = combineFrom(context, childContext);
  }
  if (isFunction(options.afterRender)) {
    options.afterRender(instance);
  }
  instance.$LI = input;
  return instance;
}
function handleComponentInput(input, componentVNode) {
  if (isInvalid(input)) {
    input = createVoidVNode();
  } else if (isStringOrNumber(input)) {
    input = createTextVNode(input, null);
  } else {
    if (input.dom) {
      input = directClone(input);
    }
    if (input.flags & 14 /* VNodeFlags.Component */) {
      // if we have an input that is also a component, we run into a tricky situation
      // where the root vNode needs to always have the correct DOM entry
      // we can optimise this in the future, but this gets us out of a lot of issues
      input.parentVNode = componentVNode;
    }
  }
  return input;
}

function mount(vNode, parentDom, context, isSVG) {
  const flags = vNode.flags;
  if (flags & 481 /* VNodeFlags.Element */) {
    return mountElement(vNode, parentDom, context, isSVG);
  }
  if (flags & 14 /* VNodeFlags.Component */) {
    return mountComponent(vNode, parentDom, context, isSVG, (flags & 4 /* VNodeFlags.ComponentClass */) > 0);
  }
  if (flags & 512 /* VNodeFlags.Void */ || flags & 16 /* VNodeFlags.Text */) {
    return mountText(vNode, parentDom);
  }
  if (flags & 1024 /* VNodeFlags.Portal */) {
    mount(vNode.children, vNode.type, context, false);
    return vNode.dom = mountText(createVoidVNode(), parentDom);
  }
}
function mountText(vNode, parentDom) {
  const dom = vNode.dom = document.createTextNode(vNode.children);
  if (!isNull(parentDom)) {
    appendChild(parentDom, dom);
  }
  return dom;
}
function mountElement(vNode, parentDom, context, isSVG) {
  const flags = vNode.flags;
  const children = vNode.children;
  const props = vNode.props;
  const className = vNode.className;
  const ref = vNode.ref;
  const childFlags = vNode.childFlags;
  isSVG = isSVG || (flags & 32 /* VNodeFlags.SvgElement */) > 0;
  const dom = documentCreateElement(vNode.type, isSVG);
  vNode.dom = dom;
  if (!isNullOrUndef(className) && className !== '') {
    if (isSVG) {
      dom.setAttribute('class', className);
    } else {
      dom.className = className;
    }
  }
  if (!isNull(parentDom)) {
    appendChild(parentDom, dom);
  }
  if ((childFlags & 1 /* ChildFlags.HasInvalidChildren */) === 0) {
    const childrenIsSVG = isSVG === true && vNode.type !== 'foreignObject';
    if (childFlags === 2 /* ChildFlags.HasVNodeChildren */) {
      mount(children, dom, context, childrenIsSVG);
    } else if (childFlags & 12 /* ChildFlags.MultipleChildren */) {
      mountArrayChildren(children, dom, context, childrenIsSVG);
    }
  }
  if (!isNull(props)) {
    mountProps(vNode, flags, props, dom, isSVG);
  }
  if (isFunction(ref)) {
    mountRef(dom, ref);
  }
  return dom;
}
function mountArrayChildren(children, dom, context, isSVG) {
  for (let i = 0, len = children.length; i < len; i++) {
    let child = children[i];
    if (!isNull(child.dom)) {
      children[i] = child = directClone(child);
    }
    mount(child, dom, context, isSVG);
  }
}
function mountComponent(vNode, parentDom, context, isSVG, isClass) {
  let dom;
  const type = vNode.type;
  const props = vNode.props || EMPTY_OBJ;
  const ref = vNode.ref;
  if (isClass) {
    const instance = createClassComponentInstance(vNode, type, props, context);
    vNode.dom = dom = mount(instance.$LI, null, instance.$CX, isSVG);
    mountClassComponentCallbacks(vNode, ref, instance);
    instance.$UPD = false;
  } else {
    const input = handleComponentInput(type(props, context), vNode);
    vNode.children = input;
    vNode.dom = dom = mount(input, null, context, isSVG);
    mountFunctionalComponentCallbacks(props, ref, dom);
  }
  if (!isNull(parentDom)) {
    appendChild(parentDom, dom);
  }
  return dom;
}
function createClassMountCallback(instance) {
  return () => {
    instance.$UPD = true;
    instance.componentDidMount();
    instance.$UPD = false;
  };
}
function mountClassComponentCallbacks(vNode, ref, instance) {
  if (isFunction(ref)) {
    ref(instance);
  }
  if (isFunction(instance.componentDidMount)) {
    LIFECYCLE.push(createClassMountCallback(instance));
  }
}
function createOnMountCallback(ref, dom, props) {
  return () => ref.onComponentDidMount(dom, props);
}
function mountFunctionalComponentCallbacks(props, ref, dom) {
  if (!isNullOrUndef(ref)) {
    if (isFunction(ref.onComponentWillMount)) {
      ref.onComponentWillMount(props);
    }
    if (isFunction(ref.onComponentDidMount)) {
      LIFECYCLE.push(createOnMountCallback(ref, dom, props));
    }
  }
}
function mountRef(dom, value) {
  LIFECYCLE.push(() => value(dom));
}

function hydrateComponent(vNode, dom, context, isSVG, isClass) {
  const type = vNode.type;
  const ref = vNode.ref;
  const props = vNode.props || EMPTY_OBJ;
  if (isClass) {
    const instance = createClassComponentInstance(vNode, type, props, context);
    const input = instance.$LI;
    hydrateVNode(input, dom, instance.$CX, isSVG);
    vNode.dom = input.dom;
    mountClassComponentCallbacks(vNode, ref, instance);
    instance.$UPD = false; // Mount finished allow going sync
  } else {
    const input = handleComponentInput(type(props, context), vNode);
    hydrateVNode(input, dom, context, isSVG);
    vNode.children = input;
    vNode.dom = input.dom;
    mountFunctionalComponentCallbacks(props, ref, dom);
  }
}
function hydrateElement(vNode, dom, context, isSVG) {
  const children = vNode.children;
  const props = vNode.props;
  const className = vNode.className;
  const flags = vNode.flags;
  const ref = vNode.ref;
  isSVG = isSVG || (flags & 32 /* VNodeFlags.SvgElement */) > 0;
  if (dom.nodeType !== 1 || dom.tagName.toLowerCase() !== vNode.type) {
    const newDom = mountElement(vNode, null, context, isSVG);
    vNode.dom = newDom;
    replaceChild(dom.parentNode, newDom, dom);
  } else {
    vNode.dom = dom;
    let childNode = dom.firstChild;
    const childFlags = vNode.childFlags;
    if ((childFlags & 1 /* ChildFlags.HasInvalidChildren */) === 0) {
      let nextSibling = null;
      while (childNode) {
        nextSibling = childNode.nextSibling;
        if (childNode.nodeType === 8) {
          if (childNode.data === '!') {
            dom.replaceChild(document.createTextNode(''), childNode);
          } else {
            dom.removeChild(childNode);
          }
        }
        childNode = nextSibling;
      }
      childNode = dom.firstChild;
      if (childFlags === 2 /* ChildFlags.HasVNodeChildren */) {
        if (isNull(childNode)) {
          mount(children, dom, context, isSVG);
        } else {
          nextSibling = childNode.nextSibling;
          hydrateVNode(children, childNode, context, isSVG);
          childNode = nextSibling;
        }
      } else if (childFlags & 12 /* ChildFlags.MultipleChildren */) {
        for (let i = 0, len = children.length; i < len; i++) {
          const child = children[i];
          if (isNull(childNode)) {
            mount(child, dom, context, isSVG);
          } else {
            nextSibling = childNode.nextSibling;
            hydrateVNode(child, childNode, context, isSVG);
            childNode = nextSibling;
          }
        }
      }
      // clear any other DOM nodes, there should be only a single entry for the root
      while (childNode) {
        nextSibling = childNode.nextSibling;
        dom.removeChild(childNode);
        childNode = nextSibling;
      }
    } else if (!isNull(dom.firstChild) && !isSamePropsInnerHTML(dom, props)) {
      dom.textContent = ''; // dom has content, but VNode has no children remove everything from DOM
      if (flags & 448 /* VNodeFlags.FormElement */) {
        // If element is form element, we need to clear defaultValue also
        dom.defaultValue = '';
      }
    }
    if (!isNull(props)) {
      mountProps(vNode, flags, props, dom, isSVG);
    }
    if (isNullOrUndef(className)) {
      if (dom.className !== '') {
        dom.removeAttribute('class');
      }
    } else if (isSVG) {
      dom.setAttribute('class', className);
    } else {
      dom.className = className;
    }
    if (isFunction(ref)) {
      mountRef(dom, ref);
    }
  }
}
function hydrateText(vNode, dom) {
  if (dom.nodeType !== 3) {
    const newDom = mountText(vNode, null);
    vNode.dom = newDom;
    replaceChild(dom.parentNode, newDom, dom);
  } else {
    const text = vNode.children;
    if (dom.nodeValue !== text) {
      dom.nodeValue = text;
    }
    vNode.dom = dom;
  }
}
function hydrateVNode(vNode, dom, context, isSVG) {
  const flags = vNode.flags;
  if (flags & 14 /* VNodeFlags.Component */) {
    hydrateComponent(vNode, dom, context, isSVG, (flags & 4 /* VNodeFlags.ComponentClass */) > 0);
  } else if (flags & 481 /* VNodeFlags.Element */) {
    hydrateElement(vNode, dom, context, isSVG);
  } else if (flags & 16 /* VNodeFlags.Text */) {
    hydrateText(vNode, dom);
  } else if (flags & 512 /* VNodeFlags.Void */) {
    vNode.dom = dom;
  } else {
    throwError();
  }
}
function hydrate(input, parentDom, callback) {
  let dom = parentDom.firstChild;
  if (!isNull(dom)) {
    if (!isInvalid(input)) {
      hydrateVNode(input, dom, EMPTY_OBJ, false);
    }
    dom = parentDom.firstChild;
    // clear any other DOM nodes, there should be only a single entry for the root
    while (dom = dom.nextSibling) {
      parentDom.removeChild(dom);
    }
  }
  if (LIFECYCLE.length > 0) {
    callAll(LIFECYCLE);
  }
  parentDom.$V = input;
  if (isFunction(callback)) {
    callback();
  }
}

function replaceWithNewNode(lastNode, nextNode, parentDom, context, isSVG) {
  unmount(lastNode);
  replaceChild(parentDom, mount(nextNode, null, context, isSVG), lastNode.dom);
}
function patch(lastVNode, nextVNode, parentDom, context, isSVG) {
  const nextFlags = nextVNode.flags | 0;
  if (lastVNode.flags !== nextFlags || nextFlags & 2048 /* VNodeFlags.ReCreate */) {
    replaceWithNewNode(lastVNode, nextVNode, parentDom, context, isSVG);
  } else if (nextFlags & 481 /* VNodeFlags.Element */) {
    patchElement(lastVNode, nextVNode, parentDom, context, isSVG, nextFlags);
  } else if (nextFlags & 14 /* VNodeFlags.Component */) {
    patchComponent(lastVNode, nextVNode, parentDom, context, isSVG, (nextFlags & 4 /* VNodeFlags.ComponentClass */) > 0);
  } else if (nextFlags & 16 /* VNodeFlags.Text */) {
    patchText(lastVNode, nextVNode);
  } else if (nextFlags & 512 /* VNodeFlags.Void */) {
    nextVNode.dom = lastVNode.dom;
  } else {
    patchPortal(lastVNode, nextVNode, context);
  }
}
function patchContentEditableChildren(dom, nextVNode) {
  if (dom.textContent !== nextVNode.children) {
    dom.textContent = nextVNode.children;
  }
}
function patchPortal(lastVNode, nextVNode, context) {
  const lastContainer = lastVNode.type;
  const nextContainer = nextVNode.type;
  const nextChildren = nextVNode.children;
  patchChildren(lastVNode.childFlags, nextVNode.childFlags, lastVNode.children, nextChildren, lastContainer, context, false);
  nextVNode.dom = lastVNode.dom;
  if (lastContainer !== nextContainer && !isInvalid(nextChildren)) {
    const node = nextChildren.dom;
    lastContainer.removeChild(node);
    nextContainer.appendChild(node);
  }
}
function patchElement(lastVNode, nextVNode, parentDom, context, isSVG, nextFlags) {
  const nextTag = nextVNode.type;
  if (lastVNode.type !== nextTag) {
    replaceWithNewNode(lastVNode, nextVNode, parentDom, context, isSVG);
  } else {
    const dom = lastVNode.dom;
    const lastProps = lastVNode.props;
    const nextProps = nextVNode.props;
    let isFormElement = false;
    let hasControlledValue = false;
    let nextPropsOrEmpty;
    nextVNode.dom = dom;
    isSVG = isSVG || (nextFlags & 32 /* VNodeFlags.SvgElement */) > 0;
    // inlined patchProps  -- starts --
    if (lastProps !== nextProps) {
      const lastPropsOrEmpty = lastProps || EMPTY_OBJ;
      nextPropsOrEmpty = nextProps || EMPTY_OBJ;
      if (nextPropsOrEmpty !== EMPTY_OBJ) {
        isFormElement = (nextFlags & 448 /* VNodeFlags.FormElement */) > 0;
        if (isFormElement) {
          hasControlledValue = isControlledFormElement(nextPropsOrEmpty);
        }
        for (const prop in nextPropsOrEmpty) {
          const lastValue = lastPropsOrEmpty[prop];
          const nextValue = nextPropsOrEmpty[prop];
          if (lastValue !== nextValue) {
            patchProp(prop, lastValue, nextValue, dom, isSVG, hasControlledValue, lastVNode);
          }
        }
      }
      if (lastPropsOrEmpty !== EMPTY_OBJ) {
        for (const prop in lastPropsOrEmpty) {
          if (!nextPropsOrEmpty.hasOwnProperty(prop) && !isNullOrUndef(lastPropsOrEmpty[prop])) {
            patchProp(prop, lastPropsOrEmpty[prop], null, dom, isSVG, hasControlledValue, lastVNode);
          }
        }
      }
    }
    const lastChildren = lastVNode.children;
    const nextChildren = nextVNode.children;
    const nextRef = nextVNode.ref;
    const lastClassName = lastVNode.className;
    const nextClassName = nextVNode.className;
    if (nextFlags & 4096 /* VNodeFlags.ContentEditable */) {
      patchContentEditableChildren(dom, nextChildren);
    } else {
      patchChildren(lastVNode.childFlags, nextVNode.childFlags, lastChildren, nextChildren, dom, context, isSVG && nextTag !== 'foreignObject');
    }
    if (isFormElement) {
      processElement(nextFlags, nextVNode, dom, nextPropsOrEmpty, false, hasControlledValue);
    }
    // inlined patchProps  -- ends --
    if (lastClassName !== nextClassName) {
      if (isNullOrUndef(nextClassName)) {
        dom.removeAttribute('class');
      } else if (isSVG) {
        dom.setAttribute('class', nextClassName);
      } else {
        dom.className = nextClassName;
      }
    }
    if (isFunction(nextRef) && lastVNode.ref !== nextRef) {
      mountRef(dom, nextRef);
    }
  }
}
function patchChildren(lastChildFlags, nextChildFlags, lastChildren, nextChildren, parentDOM, context, isSVG) {
  switch (lastChildFlags) {
    case 2 /* ChildFlags.HasVNodeChildren */:
      switch (nextChildFlags) {
        case 2 /* ChildFlags.HasVNodeChildren */:
          patch(lastChildren, nextChildren, parentDOM, context, isSVG);
          break;
        case 1 /* ChildFlags.HasInvalidChildren */:
          remove(lastChildren, parentDOM);
          break;
        default:
          remove(lastChildren, parentDOM);
          mountArrayChildren(nextChildren, parentDOM, context, isSVG);
          break;
      }
      break;
    case 1 /* ChildFlags.HasInvalidChildren */:
      switch (nextChildFlags) {
        case 2 /* ChildFlags.HasVNodeChildren */:
          mount(nextChildren, parentDOM, context, isSVG);
          break;
        case 1 /* ChildFlags.HasInvalidChildren */:
          break;
        default:
          mountArrayChildren(nextChildren, parentDOM, context, isSVG);
          break;
      }
      break;
    default:
      if (nextChildFlags & 12 /* ChildFlags.MultipleChildren */) {
        const lastLength = lastChildren.length;
        const nextLength = nextChildren.length;
        // Fast path's for both algorithms
        if (lastLength === 0) {
          if (nextLength > 0) {
            mountArrayChildren(nextChildren, parentDOM, context, isSVG);
          }
        } else if (nextLength === 0) {
          removeAllChildren(parentDOM, lastChildren);
        } else if (nextChildFlags === 8 /* ChildFlags.HasKeyedChildren */ && lastChildFlags === 8 /* ChildFlags.HasKeyedChildren */) {
          patchKeyedChildren(lastChildren, nextChildren, parentDOM, context, isSVG, lastLength, nextLength);
        } else {
          patchNonKeyedChildren(lastChildren, nextChildren, parentDOM, context, isSVG, lastLength, nextLength);
        }
      } else if (nextChildFlags === 1 /* ChildFlags.HasInvalidChildren */) {
        removeAllChildren(parentDOM, lastChildren);
      } else if (nextChildFlags === 2 /* ChildFlags.HasVNodeChildren */) {
        removeAllChildren(parentDOM, lastChildren);
        mount(nextChildren, parentDOM, context, isSVG);
      }
      break;
  }
}
function updateClassComponent(instance, nextState, nextVNode, nextProps, parentDom, context, isSVG, force, fromSetState) {
  const lastState = instance.state;
  const lastProps = instance.props;
  nextVNode.children = instance;
  let renderOutput;
  if (instance.$UN) {
    return;
  }
  if (lastProps !== nextProps || nextProps === EMPTY_OBJ) {
    if (!fromSetState && isFunction(instance.componentWillReceiveProps)) {
      instance.$BR = true;
      instance.componentWillReceiveProps(nextProps, context);
      // If instance component was removed during its own update do nothing.
      if (instance.$UN) {
        return;
      }
      instance.$BR = false;
    }
    if (instance.$PSS) {
      nextState = combineFrom(nextState, instance.$PS);
      instance.$PSS = false;
      instance.$PS = null;
    }
  }
  /* Update if scu is not defined, or it returns truthy value or force */
  const hasSCU = Boolean(instance.shouldComponentUpdate);
  if (force || !hasSCU || hasSCU && instance.shouldComponentUpdate(nextProps, nextState, context)) {
    if (isFunction(instance.componentWillUpdate)) {
      instance.$BS = true;
      instance.componentWillUpdate(nextProps, nextState, context);
      instance.$BS = false;
    }
    instance.props = nextProps;
    instance.state = nextState;
    instance.context = context;
    if (isFunction(options.beforeRender)) {
      options.beforeRender(instance);
    }
    renderOutput = instance.render(nextProps, nextState, context);
    if (isFunction(options.afterRender)) {
      options.afterRender(instance);
    }
    const didUpdate = renderOutput !== NO_OP;
    let childContext;
    if (isFunction(instance.getChildContext)) {
      childContext = instance.getChildContext();
    }
    if (isNullOrUndef(childContext)) {
      childContext = context;
    } else {
      childContext = combineFrom(context, childContext);
    }
    instance.$CX = childContext;
    if (didUpdate) {
      const lastInput = instance.$LI;
      const nextInput = handleComponentInput(renderOutput, nextVNode);
      patch(lastInput, nextInput, parentDom, childContext, isSVG);
      instance.$LI = nextInput;
      if (isFunction(instance.componentDidUpdate)) {
        instance.componentDidUpdate(lastProps, lastState);
      }
    }
  } else {
    instance.props = nextProps;
    instance.state = nextState;
    instance.context = context;
  }
  nextVNode.dom = instance.$LI.dom;
}
function patchComponent(lastVNode, nextVNode, parentDom, context, isSVG, isClass) {
  const nextType = nextVNode.type;
  const lastKey = lastVNode.key;
  const nextKey = nextVNode.key;
  if (lastVNode.type !== nextType || lastKey !== nextKey) {
    replaceWithNewNode(lastVNode, nextVNode, parentDom, context, isSVG);
  } else {
    const nextProps = nextVNode.props || EMPTY_OBJ;
    if (isClass) {
      const instance = lastVNode.children;
      instance.$UPD = true;
      instance.$V = nextVNode;
      updateClassComponent(instance, instance.state, nextVNode, nextProps, parentDom, context, isSVG, false, false);
      instance.$UPD = false;
    } else {
      let shouldUpdate = true;
      const lastProps = lastVNode.props;
      const nextHooks = nextVNode.ref;
      const nextHooksDefined = !isNullOrUndef(nextHooks);
      const lastInput = lastVNode.children;
      nextVNode.dom = lastVNode.dom;
      nextVNode.children = lastInput;
      if (nextHooksDefined && isFunction(nextHooks.onComponentShouldUpdate)) {
        shouldUpdate = nextHooks.onComponentShouldUpdate(lastProps, nextProps);
      }
      if (shouldUpdate !== false) {
        if (nextHooksDefined && isFunction(nextHooks.onComponentWillUpdate)) {
          nextHooks.onComponentWillUpdate(lastProps, nextProps);
        }
        let nextInput = nextType(nextProps, context);
        if (nextInput !== NO_OP) {
          nextInput = handleComponentInput(nextInput, nextVNode);
          patch(lastInput, nextInput, parentDom, context, isSVG);
          nextVNode.children = nextInput;
          nextVNode.dom = nextInput.dom;
          if (nextHooksDefined && isFunction(nextHooks.onComponentDidUpdate)) {
            nextHooks.onComponentDidUpdate(lastProps, nextProps);
          }
        }
      } else if (lastInput.flags & 14 /* VNodeFlags.Component */) {
        lastInput.parentVNode = nextVNode;
      }
    }
  }
}
function patchText(lastVNode, nextVNode) {
  const nextText = nextVNode.children;
  const dom = lastVNode.dom;
  if (nextText !== lastVNode.children) {
    dom.nodeValue = nextText;
  }
  nextVNode.dom = dom;
}
function patchNonKeyedChildren(lastChildren, nextChildren, dom, context, isSVG, lastChildrenLength, nextChildrenLength) {
  const commonLength = lastChildrenLength > nextChildrenLength ? nextChildrenLength : lastChildrenLength;
  let i = 0;
  let nextChild;
  let lastChild;
  for (; i < commonLength; i++) {
    nextChild = nextChildren[i];
    lastChild = lastChildren[i];
    if (nextChild.dom) {
      nextChild = nextChildren[i] = directClone(nextChild);
    }
    patch(lastChild, nextChild, dom, context, isSVG);
    lastChildren[i] = nextChild;
  }
  if (lastChildrenLength < nextChildrenLength) {
    for (i = commonLength; i < nextChildrenLength; i++) {
      nextChild = nextChildren[i];
      if (nextChild.dom) {
        nextChild = nextChildren[i] = directClone(nextChild);
      }
      mount(nextChild, dom, context, isSVG);
    }
  } else if (lastChildrenLength > nextChildrenLength) {
    for (i = commonLength; i < lastChildrenLength; i++) {
      remove(lastChildren[i], dom);
    }
  }
}
function patchKeyedChildren(a, b, dom, context, isSVG, aLength, bLength) {
  let aEnd = aLength - 1;
  let bEnd = bLength - 1;
  let i;
  let j = 0;
  let aNode = a[j];
  let bNode = b[j];
  let nextPos;
  // Step 1
  // tslint:disable-next-line
  outer: {
    // Sync nodes with the same key at the beginning.
    while (aNode.key === bNode.key) {
      if (bNode.dom) {
        b[j] = bNode = directClone(bNode);
      }
      patch(aNode, bNode, dom, context, isSVG);
      a[j] = bNode;
      j++;
      if (j > aEnd || j > bEnd) {
        break outer;
      }
      aNode = a[j];
      bNode = b[j];
    }
    aNode = a[aEnd];
    bNode = b[bEnd];
    // Sync nodes with the same key at the end.
    while (aNode.key === bNode.key) {
      if (bNode.dom) {
        b[bEnd] = bNode = directClone(bNode);
      }
      patch(aNode, bNode, dom, context, isSVG);
      a[aEnd] = bNode;
      aEnd--;
      bEnd--;
      if (j > aEnd || j > bEnd) {
        break outer;
      }
      aNode = a[aEnd];
      bNode = b[bEnd];
    }
  }
  if (j > aEnd) {
    if (j <= bEnd) {
      nextPos = bEnd + 1;
      const nextNode = nextPos < bLength ? b[nextPos].dom : null;
      while (j <= bEnd) {
        bNode = b[j];
        if (bNode.dom) {
          b[j] = bNode = directClone(bNode);
        }
        j++;
        insertOrAppend(dom, mount(bNode, null, context, isSVG), nextNode);
      }
    }
  } else if (j > bEnd) {
    while (j <= aEnd) {
      remove(a[j++], dom);
    }
  } else {
    let aStart = j;
    const bStart = j;
    const aLeft = aEnd - j + 1;
    const bLeft = bEnd - j + 1;
    const sources = [];
    for (i = 0; i < bLeft; i++) {
      sources.push(0);
    }
    // Keep track if its possible to remove whole DOM using textContent = '';
    let canRemoveWholeContent = aLeft === aLength;
    let moved = false;
    let pos = 0;
    let patched = 0;
    // When sizes are small, just loop them through
    if (bLength < 4 || (aLeft | bLeft) < 32) {
      for (i = aStart; i <= aEnd; i++) {
        aNode = a[i];
        if (patched < bLeft) {
          for (j = bStart; j <= bEnd; j++) {
            bNode = b[j];
            if (aNode.key === bNode.key) {
              sources[j - bStart] = i + 1;
              if (canRemoveWholeContent) {
                canRemoveWholeContent = false;
                while (i > aStart) {
                  remove(a[aStart++], dom);
                }
              }
              if (pos > j) {
                moved = true;
              } else {
                pos = j;
              }
              if (bNode.dom) {
                b[j] = bNode = directClone(bNode);
              }
              patch(aNode, bNode, dom, context, isSVG);
              patched++;
              break;
            }
          }
          if (!canRemoveWholeContent && j > bEnd) {
            remove(aNode, dom);
          }
        } else if (!canRemoveWholeContent) {
          remove(aNode, dom);
        }
      }
    } else {
      const keyIndex = {};
      // Map keys by their index
      for (i = bStart; i <= bEnd; i++) {
        keyIndex[b[i].key] = i;
      }
      // Try to patch same keys
      for (i = aStart; i <= aEnd; i++) {
        aNode = a[i];
        if (patched < bLeft) {
          j = keyIndex[aNode.key];
          if (j !== void 0) {
            if (canRemoveWholeContent) {
              canRemoveWholeContent = false;
              while (i > aStart) {
                remove(a[aStart++], dom);
              }
            }
            bNode = b[j];
            sources[j - bStart] = i + 1;
            if (pos > j) {
              moved = true;
            } else {
              pos = j;
            }
            if (bNode.dom) {
              b[j] = bNode = directClone(bNode);
            }
            patch(aNode, bNode, dom, context, isSVG);
            patched++;
          } else if (!canRemoveWholeContent) {
            remove(aNode, dom);
          }
        } else if (!canRemoveWholeContent) {
          remove(aNode, dom);
        }
      }
    }
    // fast-path: if nothing patched remove all old and add all new
    if (canRemoveWholeContent) {
      removeAllChildren(dom, a);
      mountArrayChildren(b, dom, context, isSVG);
    } else {
      if (moved) {
        const seq = lis_algorithm(sources);
        j = seq.length - 1;
        for (i = bLeft - 1; i >= 0; i--) {
          if (sources[i] === 0) {
            pos = i + bStart;
            bNode = b[pos];
            if (bNode.dom) {
              b[pos] = bNode = directClone(bNode);
            }
            nextPos = pos + 1;
            insertOrAppend(dom, mount(bNode, null, context, isSVG), nextPos < bLength ? b[nextPos].dom : null);
          } else if (j < 0 || i !== seq[j]) {
            pos = i + bStart;
            bNode = b[pos];
            nextPos = pos + 1;
            insertOrAppend(dom, bNode.dom, nextPos < bLength ? b[nextPos].dom : null);
          } else {
            j--;
          }
        }
      } else if (patched !== bLeft) {
        // when patched count doesn't match b length we need to insert those new ones
        // loop backwards so we can use insertBefore
        for (i = bLeft - 1; i >= 0; i--) {
          if (sources[i] === 0) {
            pos = i + bStart;
            bNode = b[pos];
            if (bNode.dom) {
              b[pos] = bNode = directClone(bNode);
            }
            nextPos = pos + 1;
            insertOrAppend(dom, mount(bNode, null, context, isSVG), nextPos < bLength ? b[nextPos].dom : null);
          }
        }
      }
    }
  }
}
// https://en.wikipedia.org/wiki/Longest_increasing_subsequence
function lis_algorithm(arr) {
  const p = arr.slice();
  const result = [0];
  let i;
  let j;
  let u;
  let v;
  let c;
  const len = arr.length;
  for (i = 0; i < len; i++) {
    const arrI = arr[i];
    if (arrI !== 0) {
      j = result[result.length - 1];
      if (arr[j] < arrI) {
        p[i] = j;
        result.push(i);
        continue;
      }
      u = 0;
      v = result.length - 1;
      while (u < v) {
        c = (u + v) / 2 | 0;
        if (arr[result[c]] < arrI) {
          u = c + 1;
        } else {
          v = c;
        }
      }
      if (arrI < arr[result[u]]) {
        if (u > 0) {
          p[i] = result[u - 1];
        }
        result[u] = i;
      }
    }
  }
  u = result.length;
  v = result[u - 1];
  while (u-- > 0) {
    result[u] = v;
    v = p[v];
  }
  return result;
}

function render(input, parentDom, callback) {
  if (input === NO_OP) {
    return;
  }
  let rootInput = parentDom.$V;
  if (isNullOrUndef(rootInput)) {
    if (!isInvalid(input)) {
      if (input.dom) {
        input = directClone(input);
      }
      if (isNull(parentDom.firstChild)) {
        mount(input, parentDom, EMPTY_OBJ, false);
        parentDom.$V = input;
      } else {
        hydrate(input, parentDom);
      }
      rootInput = input;
    }
  } else {
    if (isNullOrUndef(input)) {
      remove(rootInput, parentDom);
      parentDom.$V = null;
    } else {
      if (input.dom) {
        input = directClone(input);
      }
      patch(rootInput, input, parentDom, EMPTY_OBJ, false);
      rootInput = parentDom.$V = input;
    }
  }
  if (LIFECYCLE.length > 0) {
    callAll(LIFECYCLE);
  }
  if (isFunction(callback)) {
    callback();
  }
  if (isFunction(options.renderComplete)) {
    options.renderComplete(rootInput);
  }
  if (rootInput && rootInput.flags & 14 /* VNodeFlags.Component */) {
    return rootInput.children;
  }
}
function createRenderer(parentDom) {
  return function renderer(lastInput, nextInput) {
    if (!parentDom) {
      parentDom = lastInput;
    }
    render(nextInput, parentDom);
  };
}
function createPortal(children, container) {
  return createVNode(1024 /* VNodeFlags.Portal */, container, null, children, 0 /* ChildFlags.UnknownChildren */, null, isInvalid(children) ? null : children.key, null);
}

const resolvedPromise = typeof Promise === 'undefined' ? null : Promise.resolve();
// raf.bind(window) is needed to work around bug in IE10-IE11 strict mode (TypeError: Invalid calling object)
const fallbackMethod = typeof requestAnimationFrame === 'undefined' ? setTimeout : requestAnimationFrame.bind(window);
function nextTick(fn) {
  if (resolvedPromise) {
    return resolvedPromise.then(fn);
  }
  return fallbackMethod(fn);
}
function queueStateChanges(component, newState, callback, force) {
  if (isFunction(newState)) {
    newState = newState(component.state, component.props, component.context);
  }
  const pending = component.$PS;
  if (isNullOrUndef(pending)) {
    component.$PS = newState;
  } else {
    for (const stateKey in newState) {
      pending[stateKey] = newState[stateKey];
    }
  }
  if (!component.$PSS && !component.$BR) {
    if (!component.$UPD) {
      component.$PSS = true;
      component.$UPD = true;
      applyState(component, force, callback);
      component.$UPD = false;
    } else {
      // Async
      let queue = component.$QU;
      if (isNull(queue)) {
        queue = component.$QU = [];
        nextTick(promiseCallback(component, queue));
      }
      if (isFunction(callback)) {
        queue.push(callback);
      }
    }
  } else {
    component.$PSS = true;
    if (component.$BR && isFunction(callback)) {
      LIFECYCLE.push(callback.bind(component));
    }
  }
}
function promiseCallback(component, queue) {
  return () => {
    component.$QU = null;
    component.$UPD = true;
    applyState(component, false, () => {
      for (let i = 0, len = queue.length; i < len; i++) {
        queue[i].call(component);
      }
    });
    component.$UPD = false;
  };
}
function applyState(component, force, callback) {
  if (component.$UN) {
    return;
  }
  if (force || !component.$BR) {
    component.$PSS = false;
    const pendingState = component.$PS;
    const prevState = component.state;
    const nextState = combineFrom(prevState, pendingState);
    const props = component.props;
    const context = component.context;
    component.$PS = null;
    let vNode = component.$V;
    const lastInput = component.$LI;
    const parentDom = lastInput.dom && lastInput.dom.parentNode;
    updateClassComponent(component, nextState, vNode, props, parentDom, context, (vNode.flags & 32 /* VNodeFlags.SvgElement */) > 0, force, true);
    if (component.$UN) {
      return;
    }
    if ((component.$LI.flags & 1024 /* VNodeFlags.Portal */) === 0) {
      const dom = component.$LI.dom;
      while (!isNull(vNode = vNode.parentVNode)) {
        if ((vNode.flags & 14 /* VNodeFlags.Component */) > 0) {
          vNode.dom = dom;
        }
      }
    }
    if (LIFECYCLE.length > 0) {
      callAll(LIFECYCLE);
    }
  } else {
    component.state = component.$PS;
    component.$PS = null;
  }
  if (isFunction(callback)) {
    callback.call(component);
  }
}
class Component {
  // QUEUE
  constructor(props, context) {
    this.state = null;
    this.props = void 0;
    this.context = void 0;
    this.refs = void 0;
    // Internal properties
    this.$BR = false;
    // BLOCK RENDER
    this.$BS = true;
    // BLOCK STATE
    this.$PSS = false;
    // PENDING SET STATE
    this.$PS = null;
    // PENDING STATE (PARTIAL or FULL)
    this.$LI = null;
    // LAST INPUT
    this.$V = null;
    // VNODE
    this.$UN = false;
    // UNMOUNTED
    this.$CX = null;
    // CHILDCONTEXT
    this.$UPD = true;
    // UPDATING
    this.$QU = null;
    /** @type {object} */
    this.props = props || EMPTY_OBJ;
    /** @type {object} */
    this.context = context || EMPTY_OBJ; // context should not be mutable
  }
  forceUpdate(callback) {
    if (this.$UN) {
      return;
    }
    // Do not allow double render during force update
    queueStateChanges(this, {}, callback, true);
  }
  setState(newState, callback) {
    if (this.$UN) {
      return;
    }
    if (!this.$BS) {
      queueStateChanges(this, newState, callback, false);
    } else {
      return;
    }
  }
  // tslint:disable-next-line:no-empty
  render(_nextProps, _nextState, _nextContext) {}
}
// Public
Component.defaultProps = void 0;

var JSX = /*#__PURE__*/Object.freeze({
    __proto__: null
});

const version = "5.6.3";

export { Component, EMPTY_OBJ, JSX, NO_OP, createComponentVNode, createPortal, createRenderer, createTextVNode, createVNode, directClone, getFlagsForElementVnode, getNumberStyleValue, hydrate, linkEvent, normalizeProps, options, render, version };
